using System.Runtime.CompilerServices;
using Lagrange.Core.Internal.Event.EventArg;

namespace Lagrange.Core.Internal.Event;

public partial class EventInvoker : IDisposable
{
    private readonly Dictionary<Type, Action<EventBase>> _events;
    
    public delegate void LagrangeEvent<in TEvent>(BotContext context, TEvent e) where TEvent : EventBase;

    internal EventInvoker(BotContext context)
    {
        _events = new Dictionary<Type, Action<EventBase>>();
        
        RegisterEvent((BotOnlineEvent e) => OnBotOnlineEvent?.Invoke(context, e));
        RegisterEvent((BotOfflineEvent e) => OnBotOfflineEvent?.Invoke(context, e));
        RegisterEvent((BotLogEvent e) => OnBotLogEvent?.Invoke(context, e));
        RegisterEvent((BotCaptchaEvent e) => OnBotCaptchaEvent?.Invoke(context, e));
        RegisterEvent((GroupInvitationEvent e) => OnGroupInvitationReceived?.Invoke(context, e));
        RegisterEvent((FriendMessageEvent e) => OnFriendMessageReceived?.Invoke(context, e));
        RegisterEvent((GroupMessageEvent e) => OnGroupMessageReceived?.Invoke(context, e));
        RegisterEvent((TempMessageEvent e) => OnTempMessageReceived?.Invoke(context, e));
        RegisterEvent((GroupAdminChangedEvent e) => OnGroupAdminChangedEvent?.Invoke(context, e));
        RegisterEvent((GroupMemberIncreaseEvent e) => OnGroupMemberIncreaseEvent?.Invoke(context, e));
        RegisterEvent((GroupMemberDecreaseEvent e) => OnGroupMemberDecreaseEvent?.Invoke(context, e));
    }
    
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    private void RegisterEvent<TEvent>(Action<TEvent> action) where TEvent : EventBase => _events[typeof(TEvent)] = e => action((TEvent)e);

    internal void PostEvent(EventBase e)
    {
        Task.Run(() =>
        {
            try
            {
                if (_events.TryGetValue(e.GetType(), out var action)) action(e);
                else PostEvent(new BotLogEvent("BotContext", LogLevel.Warning, $"Event {e.GetType().Name} is not registered but pushed to invoker"));
            }
            catch (Exception ex)
            {
                PostEvent(new BotLogEvent("BotContext", LogLevel.Exception, $"{ex.StackTrace}\n{ex.Message}"));
            }
        });
    }
    
    public void Dispose()
    {
        GC.SuppressFinalize(this);
        _events.Clear();
    }
}